
# from impedance.classes.constants import *  # imports numpy as np
from edumud.file_utils import load_config


class Particle:
    def __init__(self, r, rz, eps, fi=None, hamaker=None):
        """

        Args:
            r:   radius of the colloidal particle (m)
            rz:  zeta potential in kT/e units (i.e. zeta = 25 * rel_zeta mV)
            eps: permittivity latex
            fi:  volume fraction of particles
        """
        self.radius = r
        self.rel_zeta = rz
        self.eps = eps
        self.fi = fi
        self.hamaker = hamaker
        self.beta = None  # = numpy array_like(omega)


    @classmethod
    def from_dict(cls, d):
        """ classmethod to enable constructing an instance from configuration file.
        """
        return cls(r=d["radius"], rz=d["rel_zeta"], eps=d["eps"], fi=d["fraction"], hamaker=d['Hamaker'])


def read_particles(yaml_file):
    """read particle species in colloidal system from yaml file

    Args:
        yaml_file (str):  configuration file name with .yaml extension

    Returns:
        (list): list of Particle objects

    """
    expc = load_config(yaml_file)
    part = expc['particles']        # list of dictionaries

    num_particles = len(part)
    print(f"Number of particle types: {num_particles}")
    assert(num_particles > 0)

    # make list of particles in colloid
    p_in_c = []
    [p_in_c.append(Particle(p['radius'], p['rel_zeta'], p['eps'], p['fraction'])) for p in part]

    return p_in_c


if __name__ == "__main__":
    from pathlib import Path
    DATADIR = Path(__file__).parent.parent.parent.absolute().joinpath("ImpedanceData")
    ps = read_particles(str(DATADIR / "poly.yaml"))
    print(ps[0])



import yaml


def load_config(config_name: str):
    """load configuration file.

    Args:
        config_name (str): config (yaml) file path and name

    Returns:
        (dict): yaml file contents
    """
    with open(config_name) as config_file:
        hp = yaml.safe_load(config_file)  # hp = house_parameters
        return hp




from edumud.constants import *  # imports numpy as np
from edumud.file_utils import load_config


class Electrolyte:
    def __init__(self, config_file: str):
        self.config_file = config_file
        self.C_salt = None
        self.diff_plus = None
        self.z_plus = None
        self.nu_plus = None
        self.diff_minus = None
        self.z_minus = None
        self.nu_minus = None
        self.kappa_zero2 = None
        self.kappa2 = None
        self.kappa = None
        self.D_zero = None
        self.K1 = None
        self.lambda_c2 = None
        self.lambda_n2 = None
        self.read(self.config_file)

    def calc_kappa(self):
        """calculation of kappa_zero^2
           calculation of kappa (inverse of Debije length) and kappa^2.

        """
        # calculate kappa2 / kappa: inverse of the Debye length (squared)
        self.kappa_zero2 = np.square(E_CHARGE) * self.C_salt * N_AVO
        self.kappa_zero2 /= EPS_ZW * BOLTZ_T
        kappa2 = self.kappa_zero2 * (self.nu_plus * np.square(self.z_plus) + self.nu_minus * np.square(self.z_minus))
        self.kappa2 = kappa2
        self.kappa = np.sqrt(kappa2)  # inverse of the Debije length

    def calc_D_zero(self):
        """calculate parameter D0.

        """
        self.D_zero = np.reciprocal((self.z_plus - self.z_minus) /
                                    (self.z_plus * self.diff_plus - self.z_minus * self.diff_minus))

    def calc_K1(self):
        """calculate limiting conductivity value K1.

        """
        self.K1 = EPS_ZW * self.kappa2 * self.D_zero  # limiting conductivity value

    def read(self, yaml_file: str):
        """read elektrolyte ionic species from yaml file.

        Args:
            yaml_file (str):  configuration file name with .yaml extension

        """
        expc = load_config(yaml_file)

        self.C_salt = expc['electrolyte'][0]['conc']

        self.diff_plus = expc['electrolyte'][0]['ions'][0]['D']  # ionic diff. coeff. for ion 1 ("+" ion)in m^2/s
        self.z_plus = expc['electrolyte'][0]['ions'][0]['z']  # valence ion 1 ("+" ion)
        self.nu_plus = expc['electrolyte'][0]['ions'][0]['nu']  # stoechiometric coef. ion 1 ("+" ion)

        self.diff_minus = expc['electrolyte'][0]['ions'][1]['D']  # ionic diff. coeff. for ion 2 ("-" ion)in m^2/s
        self.z_minus = expc['electrolyte'][0]['ions'][1]['z']  # valence ion 2 ("-" ion)
        self.nu_minus = expc['electrolyte'][0]['ions'][1]['nu']  # stoechiometric coef. ion 2 ("-" ion)

        neutrality = self.z_plus * self.nu_plus + self.z_minus * self.nu_minus
        print(f"Electroneutrality: {neutrality} (check if equals 0)")

        self.calc_kappa()
        self.calc_D_zero()
        self.calc_K1()           # limiting conductivity value


if __name__ == "__main__":
    from pathlib import Path
    DATADIR = Path(__file__).parent.parent.parent.absolute().joinpath("ImpedanceData")

    e = Electrolyte(str(DATADIR / "Cell_1_KCl_1mM.yaml"))
    print(f"kappa^2: {e.kappa2}")



from edumud.constants import *  # imports numpy
# from impedance.poly.polyutils_oo import complex_perm_full
from edumud.cells import Cell, PlanarCell, CylindricalCell
from edumud.electrolytes import Electrolyte


# from impedance.classes.particles import Particle, read_particles


class Colloid:
    def __init__(self, c: Cell, e: Electrolyte, part: list = None):
        self.c = c
        self.e = e
        self.part = part
        self.Cpol = None
        self.C0 = None
        self.R0 = None
        self.Rseries = None
        self.Lseries = None
        self.Z0 = None
        self.Zeq = None
        self.Keq = None
        self.EPSeq = None
        self.omega_p = None
        self.omega_c = None
        self.omega_zero = None
        self.omega_h = None

    def calc_eq_circuit(self, Rseries=0.0, Lseries=0.0):
        if isinstance(self.c, PlanarCell):
            self.Cpol = EPS_ZW * self.e.kappa * self.c.el_dist / 2.0
            self.C0 = EPS_ZW
            self.R0 = np.reciprocal(self.e.K1)
        elif isinstance(self.c, CylindricalCell):
            self.Cpol = 2.0 * np.pi * self.c.height * EPS_ZW * self.e.kappa
            self.Cpol /= self.c.one_over_inner_plus_one_over_outer

            self.C0 = 2.0 * np.pi * self.c.height * EPS_ZW
            self.C0 /= self.c.log_ratio_radii

            self.R0 = np.reciprocal(2.0 * np.pi * self.c.height)
            self.R0 *= self.c.log_ratio_radii
            self.R0 /= self.e.K1
        else:
            print("calculation of Cpol, C) and R0 not implemented for this type of cell")
        self.Rseries = Rseries  # Ohm
        self.Lseries = Lseries  # nH

        print(f"C_p = {self.Cpol * 1e6:.4g} microF")
        print(f"C_0 = {self.C0 * 1e12:.4g} pF")
        print(f"R_0 = {self.R0 / 1e3:.4g} kOhm")
        print(f"L_series = {self.Lseries * 1e9:.4g} nH")
        print(f"R_series = {self.Rseries:.4g} Ohm")

    def calc_impedance_from_eq_circuit(self, omega):
        if self.C0 is not None:
            self.Z0 = np.reciprocal(1.0 / self.R0 + 1j*self.C0*omega)  # adapt
            self.Zeq = self.Z0 + 1.0 / (1j*self.Cpol*omega)
            # multiply eq circuit with cell constant
            self.Zeq = self.Zeq * self.c.cell_constant
            # add Rseries and Lseries (they do not depend on cell constant!)
            self.Zeq += 1j * self.Lseries * omega + self.Rseries
            self.Keq = np.real(1.0 / self.Zeq)
            self.EPSeq = np.imag(1.0 / self.Zeq) / (EPS_ZERO * omega)

    def calc_omegas(self):
        """calculation of omega_p and omega_zero

        """
        if isinstance(self.c, PlanarCell):
            self.omega_p = 2.0 * self.e.D_zero * self.e.kappa / self.c.el_dist  # ang. vel. (rad/s) associated to cell
            print(f"omega_p = 2*kappa*D/L for omega = {self.omega_p:.3g} rad/s ({self.omega_p / (2 * np.pi):.3g} Hz)")
        elif isinstance(self.c, CylindricalCell):
            self.omega_c = self.e.D_zero * self.e.kappa
            self.omega_c *= self.c.one_over_inner_plus_one_over_outer  # ang. vel. (rad/s) associated to cyl. cell
            print(f"omega_c = {self.omega_c:.3g} rad/s ({self.omega_c / (2 * np.pi):.3g} Hz)")
            self.omega_h = self.e.kappa**3 * self.e.D_zero * self.c.one_over_inner_plus_one_over_outer
            print(f"omega_h = {self.omega_h:.3g} rad/s ({self.omega_h / (2 * np.pi):.3g} Hz)")
        else:
            print("calculation of omega_p not implemented for this type of cell")

        self.omega_zero = self.e.kappa2 * self.e.D_zero  # ang. vel. (rad/s) associated to Debye relaxation
        print(f"omega_0 = kappa^2 * D for omega = {self.omega_zero:.3g} rad/s ({self.omega_zero / (2 * np.pi):.3g} Hz)")


if __name__ == "__main__":
    from pathlib import Path

    DATADIR = Path(__file__).parent.parent.parent.absolute().joinpath("ImpedanceData")

    Cell_1 = PlanarCell(cc=12.5, d=4.0e-3)
    el = Electrolyte(str(DATADIR / "Cell_1_KCl_1mM.yaml"))
    # print(f"kappa^2: {e.kappa2}")
    system = Colloid(Cell_1, el)
    system.calc_omegas()
    system.calc_eq_circuit()

    w = np.logspace(1, 9, 500)
    freq = w / TWOPI
    system.calc_impedance_from_eq_circuit(w)

    # theoretical calculation for Z_e
    K_e, EPS_e, *_ = complex_perm_full(w, el, Cell_1)




import numpy as np

"""
physical constants for the edumud package.

Use:
    from edumud.constants import *
"""

EPS_WATER = 80.0
"""relative permittivity of water."""
EPS_ZERO = 8.85e-12
"""absolute permittivity of vacuum."""
EPS_ZW = EPS_ZERO * EPS_WATER  # often together
"""absolute permittivity of water EPS_ZERO* EPS_WATER."""
E_CHARGE = 1.6e-19             # elementary charge
"""elementary charge of an electron in Coulomb."""
BOLTZ_T = 1.38e-23 * 298       # Boltzmann constant x temperature
"""kT for T = 298 K."""
N_AVO = 6.0e23                 # Avogadro's number
"""Avogadro's Number (# particles in a mol)."""
ETA_VISC = 0.89e-3             # viscosity water
"""viscosity of water in SI units."""
TWOPI = 2.0 * np.pi
"""convenience constant for conversions from frequency to angular velocity."""




# import numpy as np

from edumud.constants import *  # imports numpy as np


class Cell:
    """class for general electrolysis cell.

    """
    def __init__(self, cc: float):
        """

        Args:
            cc: cell constant in m^-1
        """
        self.cell_constant = cc


class PlanarCell(Cell):
    """class for electrolysis cell with planar electrodes.

    """
    def __init__(self, cc: float, d: float):
        """

        Args:
            cc: cell constant electrode_distance/electrode surface in m^-1
            d: electrode distance in m
        """
        super().__init__(cc)
        # self.cell_constant = cc
        self.el_dist = d
        self.surface = d/cc


class CylindricalCell(Cell):
    """class for electrolysis cell with cylindrical electrodes.

    """
    def __init__(self, h: float, r_inner: float, r_outer):
        """Note: cell constant 1/(2*pi*h) surface in m^-1 in constructor of parent class "Cell".

        Args:
            h: height of cell
            r_inner: radius of cell
            r_outer: radius of cell
        """
        super().__init__(cc=np.reciprocal(2.0*np.pi*h))
        self.inner_radius = r_inner
        self.outer_radius = r_outer
        self.height = h

        self.log_ratio_radii = np.log(self.outer_radius/self.inner_radius)
        self.one_over_inner_plus_one_over_outer = np.reciprocal(self.inner_radius) + np.reciprocal(self.outer_radius)


if __name__ == "__main__":
    Cell_1 = PlanarCell(cc=12.5, d=4.0e-3)
